/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "jftpgw.h"
#include <sys/mman.h>

#define ANON_USERS " anonymous ftp "

extern int timeout;
extern struct loginstruct lsforward;
extern struct loginfo li;
extern struct hostent_list* hostcache;
static int changecode(char *const, const char*);
#ifdef SFTP_SUPPORT
static int sftp_login(struct clientinfo*);
#endif
/*static int std_login(struct clientinfo*); */

int login(struct clientinfo*, int);
static int login_connect(struct clientinfo*);
static int login_mayconnect(struct clientinfo*);
static int login_init_connection(struct clientinfo*);
static int login_connected_setup(struct clientinfo*);

static int login_auth(struct clientinfo*);
static int login_readwelcome(struct clientinfo*);
static int login_sendauth_user(struct clientinfo*);
static int login_sendauth_pass(struct clientinfo*);
static int login_successfulp(struct clientinfo*);
static int login_loggedin_setup(struct clientinfo*);


int login(struct clientinfo* clntinfo, int stage) {
	int ret = 0;
	if (stage >= LOGIN_ST_CONNECTED) {
		ret = login_connect(clntinfo);
		if (ret) { return ret; }
	}
	if (stage >= LOGIN_ST_USER) {
		ret = login_sendauth_user(clntinfo);
		if (ret) { return ret; }
	}
	if (stage >= LOGIN_ST_FULL) {
		ret = login_connect(clntinfo);
		if (ret) { return ret; }
		ret = login_auth(clntinfo);
		if (ret) { return ret; }
	}
	return 0;
}


static
int login_connect(struct clientinfo* clntinfo) {
	int ret;

	if (clntinfo->login.stage >= LOGIN_ST_CONNECTED) {
		return 0;
	}

	if ((ret = login_init_connection(clntinfo)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	if ((ret = login_connected_setup(clntinfo)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	clntinfo->login.stage = LOGIN_ST_CONNECTED;
	return 0;
}


static
int login_auth(struct clientinfo* clntinfo) {
	int ret;

	if (clntinfo->login.stage < LOGIN_ST_USER) {
		if ((ret = login_sendauth_user(clntinfo)) < 0) {
			clntinfo->login.stage = LOGIN_ST_CONNECTED;
			return ret;
		}
		clntinfo->login.stage = LOGIN_ST_USER;
	}

	if ((ret = login_sendauth_pass(clntinfo)) < 0) {
		clntinfo->login.stage = LOGIN_ST_CONNECTED;
		return ret;
	}

	if ((ret = login_loggedin_setup(clntinfo)) < 0) {
		clntinfo->login.stage = LOGIN_ST_CONNECTED;
		return ret;
	}

	return 0;
}


static
int login_mayconnect(struct clientinfo* clntinfo) {
	int allowed = 0;
	unsigned long int target;
	const char* hostname;
	char* client_ip, *server_ip;

	/* See if there was a forward and if so, check if there is a
	 * password */
	allowed = checkfwpass(clntinfo);
	if (!allowed) {
		say(clntinfo->clientsocket, "530 Login incorrect\r\n");
		return -1;
	}
	allowed = 0;

	/* we don't know yet if clntinfo->destination is a hostname like
	 * "somehost.foo.com" or if it is an IP in a char* like
	 * "212.117.232.20"
	 *
	 * So call inet_addr(), it should make the decision  :-)
	 *
	 * */
	target = inet_addr(clntinfo->destination);

	/* get the client_ip by asking information about the socket the
	 * client is connected */
	client_ip = getpeer_ip(clntinfo->clientsocket);
	server_ip = (char*) 0;

	if (target == (unsigned long int) UINT_MAX) {
		/* clntinfo->destination may be a hostname, but it needn't.
		 * It may also be an invalid IP: "393.39.239.500" or
		 * anything else. Look it up.
		 * */
		hostname = hostent_get_ip(&hostcache, clntinfo->destination);
		if (hostname) {
			/* successful lookup */
			server_ip = strdup(gethostentip(hostname));
		} else {
			/* it's not a valid IP and we could not look it up,
			 * so we could not get valid information about the
			 * destination host */
			log(7, "Nonsense destination (no IP and could not look up hostname): %s",
					clntinfo->destination);
			allowed = 0;
		}
	} else {
		/* okay, it's an IP */
		server_ip = strdup(clntinfo->destination);
	}

	if (server_ip) {
		/* we could get a valid IP, now we can evaluate if client
		 * is allowed to connect to the destination host */
		char* destuser;
		if (strcmp(lsforward.fwuser, clntinfo->user) == 0) {
			/* this was not a forward */
			/* set the argument `destuser' for
			 * is_ip_target() to 0, so that user `john'
			 * is not able to log in, if only `>john' is
			 * specified (the > allows forwards only) */
			destuser = 0;
		} else {
			destuser = clntinfo->user;
		}
		/*allowed = is_ip_target(client_ip,
				server_ip, clntinfo->destination,
				lsforward.fwuser, destuser);*/
		config_shrink_config(inet_addr(client_ip),
				inet_addr(server_ip),
				clntinfo->destination,
				lsforward.fwuser,
				destuser,
				clntinfo->destinationport,
				&hostcache);
		allowed = strcmp(config_get_option("access"), "allow") == 0;

	}

	free(client_ip);
	free(server_ip);

	if (!allowed) {
		say(clntinfo->clientsocket, "531 You are not allowed to connect to that host.\r\n");
		log(8, "Not allowed to connect to %s", clntinfo->destination);
		li.respcode = 531;
		/* free the login struct lsforward */
		free_loginst();
		return -2;
	}
	return allowed;  /* a positive int */
}


#define ERR_STR_P2 "Error connecting to %s: %s\r\n"

static
int login_init_connection(struct clientinfo* clntinfo) {
	int ss, cs = clntinfo->clientsocket;
	char* sendbuf = (char*) 0;
	int ret, err, sendbufsize;

	if ((ret = login_mayconnect(clntinfo)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	/* connect */
	ss = openportname(clntinfo->destination,
			  clntinfo->destinationport, 0, clntinfo);
	if (ss < 0) {
		err = errno;
		sendbufsize = strlen(ERR_STR_P2)
				+ strlen(clntinfo->destination)
				+ strlen(strerror(err))
				+ 1;
		sendbuf = (char*) malloc(sendbufsize);
		enough_mem(sendbuf);
		snprintf(sendbuf, sendbufsize, ERR_STR_P2,
				clntinfo->destination, strerror(err));
		say(cs, sendbuf);
		free(sendbuf);
		log(5, ERR_STR_P2, clntinfo->destination, strerror(err));
		return -1;
	}

	/* connected */
	clntinfo->serversocket = ss;

	return 0;
}
#undef ERRSTR


static
int login_connected_setup(struct clientinfo* clntinfo) {
	int ret;
	char* buffer;

	if ((ret = login_readwelcome(clntinfo)) < 0) {
		return ret;
	}

	if (clntinfo->transparent == TRANSPARENT_YES) {
		/* we are connected */
		say(clntinfo->clientsocket, clntinfo->login.welcomemsg.fullmsg);
		clntinfo->login.welcomemsg.fullmsg = (char*) 0;
		clntinfo->login.welcomemsg.lastmsg = (char*) 0;
	}

	say(clntinfo->serversocket, "SYST\r\n");
	buffer = ftp_readline(clntinfo->serversocket);
	if (!buffer) {
		if (timeout) {
			log(1, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			err_time_readline(clntinfo->clientsocket);
			return -1;
		} else {
			err_readline(clntinfo->clientsocket);
			/* -2 ?? */
			return -1;
		}
	}
	if (!strlen(buffer)) {
		log(2, "The server did not respond to the initial SYST command");
		say(clntinfo->clientsocket,
				"The server did not respond correctly\r\n");
		free(buffer);
		return -1;
	}
	free(buffer);
	buffer = (char*) 0;


	/* get the throughput rate */
	clntinfo->throughput = config_get_foption("throughput", -1.0);
		/* (lsforward.fwlogin, lsforward.login,
				inet_addr(getpeer_ip(clntinfo->clientsocket)),
				(char*) 0,
				inet_addr(getpeer_ip(clntinfo->serversocket)),
				clntinfo->destination); */

	clntinfo->serveraddr = socketinfo_get_local_ip(clntinfo->serversocket);

	return 0;
}


static
int login_readwelcome(struct clientinfo *clntinfo) {
	/* Read the welcome line */
	clntinfo->login.welcomemsg = readall(clntinfo->serversocket);
	if (!clntinfo->login.welcomemsg.fullmsg) {
		/* an error occurred */
		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			err_time_readline(clntinfo->clientsocket);
		} else {
			err_readline(clntinfo->clientsocket);
		}
		return -1;
	}

	log(9, "Connected to %s, got \"%s\" as welcome message",
			clntinfo->destination, clntinfo->login.welcomemsg.fullmsg);

	if (!checkbegin(clntinfo->login.welcomemsg.lastmsg, "220 ")) {
		log(2, "Not a valid FTP server response (%s)",
				clntinfo->login.welcomemsg.fullmsg);
		say(clntinfo->clientsocket, clntinfo->login.welcomemsg.fullmsg);
		free(clntinfo->login.welcomemsg.fullmsg);
		clntinfo->login.welcomemsg.fullmsg = (char*) 0;
		clntinfo->login.welcomemsg.lastmsg = (char*) 0;
		return -1;
	}
	return 0;
}


static
int login_sendauth_user(struct clientinfo* clntinfo) {
	int sendbufsize, ret;
	char* sendbuf;

	if (clntinfo->login.stage >= LOGIN_ST_USER) {
		return 0;
	}

	sendbufsize = strlen("USER \r\n") + strlen(clntinfo->user) + 1;
	sendbuf = (char*) malloc(sendbufsize);
	enough_mem(sendbuf);

	snprintf(sendbuf, sendbufsize, "USER %s\r\n", clntinfo->user);
	ret = say(clntinfo->serversocket, sendbuf);
	free(sendbuf);
	sendbuf = 0;
	if (ret < 0) {
		log(2, "Error writing the user name to the server: %s",
			strerror(errno));
		return -1;
	}

	clntinfo->login.authresp = readall(clntinfo->serversocket);

	if (!clntinfo->login.authresp.fullmsg && timeout) {
		log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
		err_time_readline(clntinfo->clientsocket);
		free(clntinfo->login.welcomemsg.fullmsg);
		return -1;
	}
	if (!clntinfo->login.authresp.fullmsg) {
		err_readline(clntinfo->clientsocket);
		free(clntinfo->login.welcomemsg.fullmsg);
		return -1;
	}

	clntinfo->login.stage = LOGIN_ST_USER;
	return 0;
}


static
int login_sendauth_pass(struct clientinfo* clntinfo) {
	int sendbufsize, ret;
	char* sendbuf, *buffer;

	buffer = clntinfo->login.authresp.lastmsg;
	if (!checkdigits(buffer, 331) && !checkdigits(buffer, 230)) {
		/* prepend the response from the server with an error code
		 * and return it to the client */
		sendbufsize = strlen(buffer) + 4 + 1;
		sendbuf = (char*) malloc(sendbufsize);
		enough_mem(sendbuf);
		snprintf(sendbuf, sendbufsize, "500 %s", buffer);
		say(clntinfo->clientsocket, sendbuf);
		free(sendbuf);

		log(7, "Got \"%s\" after sending the username.", buffer);
		free(clntinfo->login.authresp.fullmsg);
		free(clntinfo->login.welcomemsg.fullmsg);
		return -1;
	}

	if (!checkdigits(buffer, 230)) {
		char* userdup;
		int size;

		sendbufsize = strlen("PASS \r\n") + strlen(clntinfo->pass) + 1;
		sendbuf = (char*) malloc(sendbufsize);
		enough_mem(sendbuf);
		mlock(sendbuf, sendbufsize);
		snprintf(sendbuf, sendbufsize, "PASS %s\r\n", clntinfo->pass);
		ret = say(clntinfo->serversocket, sendbuf);

		size = strlen(clntinfo->user) + 3;
		userdup = (char*) malloc( size );
		snprintf(userdup, size, " %s ", clntinfo->user);

		if (strstr(ANON_USERS, userdup)) {
			clntinfo->anon_user = clntinfo->pass;
			clntinfo->pass = (char*) 0;
			munlock(sendbuf, sendbufsize);
		} else {
			memset(sendbuf, (char) 0, sendbufsize);
			free(sendbuf);
			clntinfo->anon_user = strdup("UNKNOWN");
		}
		free(userdup);
		sendbuf = userdup = (char*) 0;
		if (ret < 0) {
			log(2, "Error writing the password to the server: %s",
				strerror(errno));
			free(clntinfo->login.welcomemsg.fullmsg);
			free(clntinfo->login.authresp.fullmsg);
			return -1;
		}
		free(clntinfo->login.authresp.fullmsg);
	}
	return 0;
}


static
int login_successfulp(struct clientinfo* clntinfo) {
	int bufsize;
	char* buffer;
	char codebuf[5];

	clntinfo->login.authresp = readall(clntinfo->serversocket);
	if (!clntinfo->login.authresp.fullmsg) {
		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,
					__LINE__); 
			err_time_readline(clntinfo->clientsocket);
		}
		else {
			err_readline(clntinfo->clientsocket);
		}
		if (clntinfo->login.welcomemsg.fullmsg) {
			free(clntinfo->login.welcomemsg.fullmsg);
		}
		return -1;
	}

	if (clntinfo->login.welcomemsg.fullmsg) {
		/* we have not yet sent the welcome message to the client */
		bufsize = strlen(clntinfo->login.welcomemsg.fullmsg)
			  + strlen(clntinfo->login.authresp.fullmsg) + 1;
		buffer = (char*) malloc(bufsize);
		enough_mem(buffer);

		/* change the status code of the welcomemsg */

		/* save the digit code of the server response from the
		 * authentication in codebuf */
		strncpy(codebuf, clntinfo->login.authresp.fullmsg, 4);
		codebuf[3] = '-';
		codebuf[4] = '\0';

		/* Change the digit code of the welcome message to the one
		 * of the server response from the authentication */
		changecode(clntinfo->login.welcomemsg.fullmsg, codebuf);
		snprintf(buffer, bufsize, "%s%s",
				clntinfo->login.welcomemsg.fullmsg,
				clntinfo->login.authresp.fullmsg);
	} else {
		buffer = strdup(clntinfo->login.authresp.fullmsg);
		enough_mem(buffer);
	}

	li.respcode = getcode(clntinfo->login.authresp.lastmsg);

	if (clntinfo->login.welcomemsg.fullmsg) {
		free(clntinfo->login.welcomemsg.fullmsg);
		clntinfo->login.welcomemsg.fullmsg 
			= clntinfo->login.welcomemsg.lastmsg 
			= (char*) 0;
	}

	if (!checkdigits(clntinfo->login.authresp.lastmsg, 230)) {
		say(clntinfo->clientsocket, clntinfo->login.authresp.lastmsg);
		log(8, "Got \"%s\" after sending the password",
				clntinfo->login.authresp.fullmsg);
		free(clntinfo->login.authresp.fullmsg);
		return -1;
	}

	/* Now the client receives the welcome message as well as the
	 * response of the server from the authentication with the same code
	 * at the beginning */
	say(clntinfo->clientsocket, buffer);
	free(buffer);
	buffer =0;

	/* free the login struct lsforward in *_cmds.c after sending */
	/* the login message */
	free_loginst();

	/* successful login */
	log(6, "Logged in to %s as %s!",
		clntinfo->destination, clntinfo->user);

	return 0;
}


static
int login_loggedin_setup(struct clientinfo* clntinfo) {
	struct in_addr iaddr;

	/* initialize the loginfo structure with the values that have to be
	 * set only once after a successful login */
	li.anon_user = clntinfo->anon_user;
	li.svrip = getpeer_ip(clntinfo->serversocket);
	li.svrname = hostent_get_name(&hostcache, inet_addr(li.svrip));
	iaddr.s_addr = clntinfo->serveraddr;
	li.ifipsvr = strdup(inet_ntoa(iaddr));
	enough_mem(li.ifipsvr);
	li.svrlogin = clntinfo->destination;

	return login_successfulp(clntinfo);
}


/* login handles the login to the server (the data is passed via CLNTINFO)
 *
 * Parameters: clntinfo: connection information
 * 
 * Return values: -1 for a system call failure (so that it can be
 *                   written via strerror to the client
 *                -2 for other errors
 *                 0 on success
 * Called by: handlecmds() after that version has read the username and
 *            maybe the password
 *
 *
 * Login response codes: 
 * 	Success:
 *		- forwards server response
 * 	Error:
 * 		- 530: Login incorrect
 * 		- 531: You are not allowed to connect to that host
 * 		- 532: Could not look up the destination host
 *		- 533: Error connecting to the server via ssh
 */





/*

	switch(clntinfo->servermode) {
#ifdef SFTP_SUPPORT
		case SECURE_SFTP:
			ret = sftp_login(clntinfo);
			break;
#endif
		default:
			ret = std_login(clntinfo);
	}

	if (ret) {
		return ret;
	}
*/




#ifdef SFTP_SUPPORT

static
int sftp_login(struct clientinfo* clntinfo) {

	int ret;
	char* args[6] = { "progname", "-l", clntinfo->user, "", "", "" };
	const char* crypt_pw;
	const char* ssh_options;
	int pw_match = 0;
	int nargs;
	struct in_addr in;

	if (!gethostaddr(clntinfo->destination,
				&in)) {
		say(clntinfo->clientsocket, "532 Could not look up the "
				"destination host\r\n");
		li.respcode = 532;
		return -2;
	}
	crypt_pw = get_sshpasswd(in.s_addr,
			clntinfo->user);
	if ( ! crypt_pw) {
		pw_match = 0;
		log(3, "No password specified for user %s",
				clntinfo->user);
	} else {
		pw_match = (cryptcmp(crypt_pw,
			clntinfo->pass) == 0);
	}
	if (!pw_match) {
		say(clntinfo->clientsocket, "530 Login incorrect.\r\n");
		li.respcode = 530;
		free(clntinfo->user);
		clntinfo->user =0;
		return -2;
	}

	ssh_options = get_sshoptions(in.s_addr, clntinfo->user);
	if (ssh_options) {
		args[3] = "-s";
		args[4] = strdup(ssh_options);
		args[5] = clntinfo->destination;
		enough_mem(args[5]);
		nargs = 6;
	} else {
		args[3] = clntinfo->destination;
		nargs = 4;
	}

	ret = sftp_open(args, nargs);
	if (nargs == 6) {
		free(args[4]);
	}
	if (ret) {
		log(2, "Error connecting to the sftp server");
		say(clntinfo->clientsocket, "533 Error connecting to the server via ssh\r\n"); 
		li.respcode = 533;
		return -2;
	}
	return 0;
}

#endif /* SFTP_SUPPORT */



/* changecode() changes the response code at the beginning of each line to a
 * new value. In fact, NEW is a _char*_, so you can put any data at the
 * beginning of each line, not just a number
 * 
 * Parameters: msg: the origianl message
 *             new: the portion that should be prepended.
 *
 * Return values: 0 on success
 *
 * Called by login() in order to display the original 220 welcome message
 *           after the login with a changed code.
 * 
 * */

static int changecode(char *const msg, const char* new) {
	char *nextline = msg;
	while (nextline) {
		if (nextline[0] >= '0' && nextline[0] <= '9') {
			/* we subtract 2 because of "\r\n". If nextline is
			 * "foo\r\n" we may at most substitute "bar" or
			 * another 3 character long string, since 
			 *      3       ==        5         - 2  */
			if (strlen(new) <= strlen(nextline) - 2) {
				strncpy(nextline, new, strlen(new));
			}
		}
		nextline = strstr(nextline, "\r\n");
		nextline += 2;
		if (!*nextline) { /* end */
			nextline = 0;
		}
	}
	return 0;
}




void free_loginst(void) {
	if (lsforward.fwlogin) {
		free(lsforward.fwlogin);
	}
	if (lsforward.fwpass) {
		free(lsforward.fwpass);
	}
	/* fwuser is always used */
	free(lsforward.fwuser);
	if (lsforward.login) {
		free(lsforward.login);
	}
	if (lsforward.pass) {
		free(lsforward.pass);
	}
	lsforward.fwlogin = lsforward.fwpass = lsforward.fwuser
		= lsforward.login = lsforward.pass = 0;
}


