/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <syslog.h>
#include <stdarg.h>
#include <pwd.h>
#include <sys/types.h>
#include "jftpgw.h"


/* from jftpgw.c */
extern char* conffilename;

struct hostent_list* hostcache;
struct uidstruct runasuser;

/*struct portrangestruct *actvrangebase, *pasvrangebase,
			*actvrangecur, *pasvrangecur;
unsigned int actvportsamount;
unsigned int pasvportsamount;
*/

struct loginstruct lsforward;
struct limitstruct *limitbase, *limitcur;


int is_below_limit(unsigned long int iaddr, struct limitstruct** ls) {

	int ret;

	if (!limitbase) {
		/* no limit options specified */
		*ls = 0;
		return 1;
	}
	limitcur = limitbase;

	/* lets see if there is a rule for the source */

	do {
		if (limitcur->hostname == NULL) {
			/* compare IPs */
			ret = match_addrs(iaddr, (const char*) 0,
					limitcur->ip.ip, limitcur->ip.netmask,
					(const char*) 0);
		} else {
			/* compare hostnames */
			ret = match_addrs(iaddr, (const char*) 0,
					-1, -1,
					limitcur->hostname);
		}
		if (ret) {
			/* found a match */
			break;
		}
	} while ((limitcur = limitcur->next));

	if (!ret) {
		/* no match found, the client may come in */
		*ls = 0;
		return 1; 
	}

	/* okay, lets see if we fit in */

	if (limitcur->connected == limitcur->connmax) {
		/* ooops... */
		*ls = 0;
		return 0;
	}

	*ls = limitcur;
	return 1;
}


int register_pid(struct limitstruct* ls, pid_t pid) {

	struct connliststruct *cls, *tmp;

	/* well, let me in */

	/* maybe we don't have an entry for that source */
	if (!ls) {
		return 0;
	}

	cls = (struct connliststruct*) malloc(sizeof(struct connliststruct));
	enough_mem(cls);

	if (!ls->connlist) {
		ls->connlist = cls;
	} else {
		tmp = ls->connlist;
		while (tmp->next) {
			tmp = tmp->next;
		}
		tmp->next = cls;
	}
	cls->next = 0;
	cls->pid = pid;
	log(9, "Added pid %d", cls->pid);
	ls->connected++;

	return 0;
}

int unregister_pid(pid_t pid) {

	struct connliststruct* cls, *prev = 0;

	if (!limitbase) {
		/* no limit options specified */
		/* or we are in a fork()ed server
		 * NOTE: Only the server that listens on the port and
		 * fork()s at every connnection attempt may have a
		 * limitbase != 0  */
		return 0;
	}

	limitcur = limitbase;

	do {
		cls = limitcur->connlist;
		while (cls) {
			log(8, "checking %d against %d", cls->pid, pid);
			if (cls->pid == pid) {
				if (prev) {
					prev->next = cls->next;
				} else {
					limitcur->connlist = cls->next;
				}
				log(9, "removed a connlist structure entry");
				limitcur->connected--;
				free(cls);
				return 0;
			}
			prev = cls;
			cls = cls->next;
		}
	} while ((limitcur = limitcur->next));

	/* pid was not found. Maybe we don't have a limit entry for that
	 * source */

	return 0;
}


int get_forward(const char* user, struct loginstruct* logininfo) {

	struct slist_t* forward_list, *fw_list_cur;
	const char* option;
	/*
	 * <user ftp>
	 *    forward johnfred@fooserver.com
	 *    forward johnfred@fooserver.com,3949,p
	 *    forward johnfred@fooserver.com     *     johnspass
	 *    forward johnfred@fooserver.com   JiKe94  johnspass
	 * </user>
	 */

	if ( ! (option = config_get_option("forward")) ) {
		memset(logininfo, 0, sizeof(struct loginstruct));
		return 0;
	}
	if ( ! (forward_list = config_split_line( option, WHITESPACES )) ) {
		memset(logininfo, 0, sizeof(struct loginstruct));
		return 0;
	}

	fw_list_cur = forward_list;

	logininfo->fwlogin = strdup(user);
	enough_mem(logininfo->fwlogin);

	logininfo->login = strdup(fw_list_cur->value);
	enough_mem(logininfo->login);

	fw_list_cur = fw_list_cur->next;

	if (fw_list_cur->value) {
		logininfo->fwpass = strdup(fw_list_cur->value);
		enough_mem(logininfo->fwpass);
	} else {
		logininfo->fwpass = 0;
	}

	fw_list_cur = fw_list_cur->next;
	if (fw_list_cur->value) {
		logininfo->pass = strdup(fw_list_cur->value);
		enough_mem(logininfo->pass);
	} else {
		logininfo->pass = 0;
	}

	slist_destroy(forward_list);
	return 1;
}


int set_conffilename(const char* arg) {
	/* TEST */
	const int increment = 1;
	int pathsize = 1;
	char* ret;
	char* path = (char*) malloc(pathsize);
	char* fname;

	enough_mem(path);
	free(conffilename);
	while ( ! (ret = getcwd(path, pathsize)) ) {
		pathsize += increment;
		if (pathsize > PATH_MAX) {
			conffilename = (char*) 0;
			free(path);
			return -1;
		}
		path = (char*) realloc(path, pathsize);
	}

	fname = (char*) malloc(PATH_MAX + 1);
	conffilename = rel2abs(arg, path, fname, PATH_MAX);
	if ( ! conffilename ) {
		free(path); free(fname);
		return -1;
	} else {
		/* reduce allocated memory */
		conffilename = realloc(fname, strlen(fname) + 1);
		enough_mem(conffilename);
		free(path);
	}
	log(9, "Conffile set to %s", conffilename);
	return 0;
}

void free_cmdlogentst(struct cmdlogent_t* cls, struct cmdlogent_t* base) {
	if (!cls) {
		return;
	}
	free_cmdlogentst(cls->next, base);
	if (cls->logf_name) {
		free(cls->logf_name);
	}
	if (cls->specs) {
		free(cls->specs);
	}
	if (cls->logf) {
		fclose(cls->logf);
		cls->logf = (FILE*) 0;
	}
	if (cls->style) {
		free(cls->style);
	}
	if (cls == base) {
		cls->logf_name = (char*) 0;
		cls->specs = (char*) 0;
	} else {
		free(cls);
	}
}

void reset_confst(struct configstruct* cs) {
	if (!cs) {
		log(1, "configstruct* cs was NULL, this should NEVER happen!");
		return;
	}
	if (cs->logf_name) {
		free(cs->logf_name);
		cs->logf_name = (char*) 0;
	}
	if (cs->logf) {
		fclose(cs->logf);
		cs->logf = (FILE*) 0;
	}
	if (cs->syslog) {
		closelog();
	}
	free_cmdlogentst(&cs->cmdlogfiles, &cs->cmdlogfiles);
	free_cmdlogentst(&cs->cmdlogdirs, &cs->cmdlogdirs);
	cs->cmdlogfiles.next = (struct cmdlogent_t*) 0;
	cs->cmdlogdirs.next = (struct cmdlogent_t*) 0;
}


/* this is called by every client at the beginning, only the main process
   keeps track of the logins
*/

void free_limitst(struct limitstruct* ls) {
	if (!ls) {
		return;
	}
	free_limitst(ls->next);
	if (ls->hostname) { free(ls->hostname); }
	free(ls);
}


int getservermode() {
	int servmode = ASCLIENT;
	const char* opt = config_get_option("defaultmode");

	/* opt should be set in every case */
	if (opt) {
		if (0 == strcasecmp(opt, "passive")) {
			servmode = PASSIVE;
		}
		if (0 == strcasecmp(opt, "active")) {
			servmode = ACTIVE;
		}
		if (0 == strcasecmp(opt, "asclient")) {
			servmode = ASCLIENT;
		}
		if (0 == strcasecmp(opt, "secure-sftp")) {
			servmode = SECURE_SFTP;
		}
	}
	return servmode;
}


int checkfwpass(struct clientinfo* clntinfo) {
	int fw = lsforward.login != 0;
	if (fw == 0) {
		/* no forward for this user. Allow further processing */
		return 1;
	}
	if (!lsforward.fwpass) {
		/* no password */
		return 1;
	} 
	if (strcmp(lsforward.fwpass, "*") == 0
		|| cryptcmp(lsforward.fwpass, clntinfo->pass) == 0) {
		/* allow any pw but use destpw on dest */
		memset(clntinfo->pass, 0, strlen(clntinfo->pass));
		clntinfo->pass = realloc(clntinfo->pass,
					strlen(lsforward.pass) + 1);
		enough_mem(clntinfo->pass);
		strncpy(clntinfo->pass, lsforward.pass,
				strlen(lsforward.pass) /* always true */);
		return 1;
	}
	return 0;
}


/*
 * Takes a socket descriptor and returns the string contain the peer's
 * IP address.  -- from tinyproxy
 */
char *getpeer_ip(int fd)
{
	const int ipaddrsize = 16;
	char* ipaddr = (char*) malloc(ipaddrsize);
	struct sockaddr_in name;
#ifdef HAVE_SOCKLEN_T
	socklen_t namelen;
#else
	int namelen;
#endif
	namelen = sizeof(name);
	
	if (getpeername(fd, (struct sockaddr *) &name, &namelen) != 0) {
		log(2, "Could not get peername: %s", strerror(errno));
		strcpy(ipaddr, "255.255.255.255");
	} else{
		strncpy(ipaddr, inet_ntoa(*(struct in_addr *) &name.sin_addr.s_addr), ipaddrsize);
	}

	return ipaddr;
}


unsigned int get_usesport(int fd) {
	struct sockaddr_in sin;
#ifdef HAVE_SOCKLEN_T
        socklen_t slen;
#else
        int slen;
#endif

	slen = sizeof(struct sockaddr);
	if (getsockname(fd, (struct sockaddr*) &sin, &slen) < 0) {
		log(2, "getsockname failed trying to get the port number: %s",
				strerror(errno));
		return DEFAULTBINDPORT;
	}
	return ntohs(sin.sin_port);
}


int save_runasuser_uid(void) {
	const char* option;
	struct passwd* pws;

	option = config_get_option("runasuser");
	if (option) {
		runasuser.username = strdup(option);
		enough_mem(runasuser.username);
		runasuser.username = trim(runasuser.username);
		pws = getpwnam(runasuser.username);
		if (!pws) {
			log(3, "getpwnam(2) could not look up user %s: %s",
				runasuser.username, strerror(errno));
			perror("Could not look up the user name");
			return -1;
		}
		runasuser.uid = pws->pw_uid;
	}
	return 0;
}

